@@ -62,7 +62,7 @@
  * 06.11.96  - Added v2 code submitted by Joerg Dorchain, but written by
  *             Andreas Schwab.
  *
- * 1999-02-22 Arkadiusz Mikiewicz <misiek@misiek.eu.org>
+ * 1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
  * - added Native Language Support
  *
  *
@@ -99,10 +99,6 @@
 #include "minix.h"
 #include "nls.h"
 
-#ifdef MINIX2_SUPER_MAGIC2
-#define HAVE_MINIX2 1
-#endif
-
 #ifndef __linux__
 #define volatile
 #endif
@@ -111,13 +107,9 @@
 
 #define UPPER(size,n) ((size+((n)-1))/(n))
 #define INODE_SIZE (sizeof(struct minix_inode))
-#ifdef HAVE_MINIX2
 #define INODE_SIZE2 (sizeof(struct minix2_inode))
 #define INODE_BLOCKS UPPER(INODES, (version2 ? MINIX2_INODES_PER_BLOCK \
 				    : MINIX_INODES_PER_BLOCK))
-#else
-#define INODE_BLOCKS UPPER(INODES, (MINIX_INODES_PER_BLOCK))
-#endif
 #define INODE_BUFFER_SIZE (INODE_BLOCKS * BLOCK_SIZE)
 
 #define BITS_PER_BLOCK (BLOCK_SIZE<<3)
@@ -142,6 +134,9 @@ static int termios_set = 0;
 #define MAX_DEPTH 50
 static int name_depth = 0;
 static char name_list[MAX_DEPTH][NAME_MAX+1];
+/* Copy of the previous, just for error reporting - see get_current_name */
+/* This is a waste of 12kB or so. */
+static char current_name[MAX_DEPTH*(NAME_MAX+1)+1];
 
 static char * inode_buffer = NULL;
 #define Inode (((struct minix_inode *) inode_buffer)-1)
@@ -149,11 +144,7 @@ static char * inode_buffer = NULL;
 static char super_block_buffer[BLOCK_SIZE];
 #define Super (*(struct minix_super_block *)super_block_buffer)
 #define INODES ((unsigned long)Super.s_ninodes)
-#ifdef HAVE_MINIX2
 #define ZONES ((unsigned long)(version2 ? Super.s_zones : Super.s_nzones))
-#else
-#define ZONES ((unsigned long)(Super.s_nzones))
-#endif
 #define IMAPS ((unsigned long)Super.s_imap_blocks)
 #define ZMAPS ((unsigned long)Super.s_zmap_blocks)
 #define FIRSTZONE ((unsigned long)Super.s_firstdatazone)
@@ -169,9 +160,7 @@ static unsigned char * inode_count = NULL;
 static unsigned char * zone_count = NULL;
 
 static void recursive_check(unsigned int ino);
-#ifdef HAVE_MINIX2
 static void recursive_check2(unsigned int ino);
-#endif
 
 #include "bitops.h"
 
@@ -210,13 +199,21 @@ die(const char *str) {
  * current file.
  */
 static void
-print_current_name(void) {
-	int i = 0;
+get_current_name(void) {
+	int i = 0, ct;
+	char *p, *q;
 
-	while (i < name_depth)
-		printf("/%.*s", namelen, name_list[i++]);
+	q = current_name;
+	while (i < name_depth) {
+		p = name_list[i++];
+		ct = namelen;
+		*q++ = '/';
+		while (ct-- && *p)
+			*q++ = *p++;
+	}
 	if (i == 0)
-		printf ("/");
+		*q++ = '/';
+	*q = 0;
 }
 
 static int
@@ -315,14 +312,18 @@ static int
 check_zone_nr(unsigned short * nr, int * corrected) {
 	if (!*nr)
 		return 0;
-	if (*nr < FIRSTZONE)
-		printf(_("Zone nr < FIRSTZONE in file `"));
-	else if (*nr >= ZONES)
-		printf(_("Zone nr >= ZONES in file `"));
-	else
+
+	if (*nr < FIRSTZONE) {
+		get_current_name();
+		printf(_("Zone nr < FIRSTZONE in file `%s'."),
+		       current_name);
+	} else if (*nr >= ZONES) {
+		get_current_name();
+		printf(_("Zone nr >= ZONES in file `%s'."),
+		       current_name);
+	} else
 		return *nr;
-	print_current_name();
-	printf("'.");
+
 	if (ask(_("Remove block"),1)) {
 		*nr = 0;
 		*corrected = 1;
@@ -330,26 +331,28 @@ check_zone_nr(unsigned short * nr, int * corrected) {
 	return 0;
 }
 
-#ifdef HAVE_MINIX2
 static int
 check_zone_nr2 (unsigned int *nr, int *corrected) {
 	if (!*nr)
 		return 0;
-	if (*nr < FIRSTZONE)
-		printf (_("Zone nr < FIRSTZONE in file `"));
-	else if (*nr >= ZONES)
-		printf (_("Zone nr >= ZONES in file `"));
-	else
+
+	if (*nr < FIRSTZONE) {
+		get_current_name();
+		printf (_("Zone nr < FIRSTZONE in file `%s'."),
+			current_name);
+	} else if (*nr >= ZONES) {
+		get_current_name();
+		printf (_("Zone nr >= ZONES in file `%s'."),
+			current_name);
+	} else
 		return *nr;
-	print_current_name ();
-	printf ("'.");
+
 	if (ask (_("Remove block"), 1)) {
 		*nr = 0;
 		*corrected = 1;
 	}
 	return 0;
 }
-#endif
 
 /*
  * read-block reads block nr into the buffer at addr.
@@ -361,15 +364,15 @@ read_block(unsigned int nr, char * addr) {
 		return;
 	}
 	if (BLOCK_SIZE*nr != lseek(IN, BLOCK_SIZE*nr, SEEK_SET)) {
-		printf(_("Read error: unable to seek to block in file '"));
-		print_current_name();
-		printf("'\n");
+		get_current_name();
+		printf(_("Read error: unable to seek to block in file '%s'\n"),
+		       current_name);
 		memset(addr,0,BLOCK_SIZE);
 		errors_uncorrected = 1;
 	} else if (BLOCK_SIZE != read(IN, addr, BLOCK_SIZE)) {
-		printf(_("Read error: bad block in file '"));
-		print_current_name();
-		printf("'\n");
+		get_current_name();
+		printf(_("Read error: bad block in file '%s'\n"),
+		       current_name);
 		memset(addr,0,BLOCK_SIZE);
 		errors_uncorrected = 1;
 	}
@@ -391,9 +394,9 @@ write_block(unsigned int nr, char * addr) {
 	if (BLOCK_SIZE*nr != lseek(IN, BLOCK_SIZE*nr, SEEK_SET))
 		die(_("seek failed in write_block"));
 	if (BLOCK_SIZE != write(IN, addr, BLOCK_SIZE)) {
-		printf(_("Write error: bad block in file '"));
-		print_current_name();
-		printf("'\n");
+		get_current_name();
+		printf(_("Write error: bad block in file '%s'\n"),
+		       current_name);
 		errors_uncorrected = 1;
 	}
 }
@@ -437,7 +440,6 @@ map_block(struct minix_inode * inode, unsigned int blknr) {
 	return result;
 }
 
-#ifdef HAVE_MINIX2
 static int
 map_block2 (struct minix2_inode *inode, unsigned int blknr) {
   	unsigned int ind[BLOCK_SIZE >> 2];
@@ -494,7 +496,6 @@ map_block2 (struct minix2_inode *inode, unsigned int blknr) {
 		write_block (block, (char *) ind);
 	return result;
 }
-#endif
 
 static void
 write_super_block(void) {
@@ -535,11 +536,9 @@ get_dirsize (void) {
 	char blk[BLOCK_SIZE];
 	int size;
 
-#if HAVE_MINIX2
 	if (version2)
 		block = Inode2[ROOT_INO].i_zone[0];
 	else
-#endif
 		block = Inode[ROOT_INO].i_zone[0];
 	read_block (block, blk);
 	for (size = 16; size < BLOCK_SIZE; size <<= 1) {
@@ -566,7 +565,6 @@ read_superblock(void) {
 		namelen = 30;
 		dirsize = 32;
 		version2 = 0;
-#ifdef HAVE_MINIX2
 	} else if (MAGIC == MINIX2_SUPER_MAGIC) {
 		namelen = 14;
 		dirsize = 16;
@@ -575,7 +573,6 @@ read_superblock(void) {
 		namelen = 30;
 		dirsize = 32;
 		version2 = 1;
-#endif
 	} else
 		die(_("bad magic number in super-block"));
 	if (ZONESIZE != 0 || BLOCK_SIZE != 1024)
@@ -637,10 +634,10 @@ get_inode(unsigned int nr) {
 	inode = Inode + nr;
 	if (!inode_count[nr]) {
 		if (!inode_in_use(nr)) {
-			printf(_("Inode %d marked not used, but used for file '"),
-				nr);
-			print_current_name();
-			printf("'\n");
+			get_current_name();
+			printf(_("Inode %d marked unused, "
+				 "but used for file '%s'\n"),
+			       nr, current_name);
 			if (repair) {
 				if (ask(_("Mark in use"),1))
 					mark_inode(nr);
@@ -663,8 +660,9 @@ get_inode(unsigned int nr) {
 		else if (S_ISFIFO(inode->i_mode))
 			;
 		else {
-                        print_current_name();
-                        printf(_(" has mode %05o\n"),inode->i_mode);
+                        get_current_name();
+                        printf(_("The file `%s' has mode %05o\n"),
+			       current_name, inode->i_mode);
                 }
 
 	} else
@@ -677,7 +675,6 @@ get_inode(unsigned int nr) {
 	return inode;
 }
 
-#ifdef HAVE_MINIX2
 static struct minix2_inode *
 get_inode2 (unsigned int nr) {
 	struct minix2_inode *inode;
@@ -688,9 +685,10 @@ get_inode2 (unsigned int nr) {
 	inode = Inode2 + nr;
 	if (!inode_count[nr]) {
 		if (!inode_in_use (nr)) {
-			printf (_("Inode %d marked not used, but used for file '"), nr);
-			print_current_name ();
-			printf ("'\n");
+			get_current_name();
+			printf (_("Inode %d marked unused, "
+				  "but used for file '%s'\n"),
+				nr, current_name);
 			if (repair) {
 				if (ask (_("Mark in use"), 1))
 					mark_inode (nr);
@@ -711,8 +709,9 @@ get_inode2 (unsigned int nr) {
 		else if (S_ISSOCK (inode->i_mode));
 		else if (S_ISFIFO (inode->i_mode));
 		else {
-			print_current_name ();
-			printf (_(" has mode %05o\n"), inode->i_mode);
+			get_current_name ();
+			printf (_("The file `%s' has mode %05o\n"),
+				current_name, inode->i_mode);
 		}
 	} else
 		links++;
@@ -723,7 +722,6 @@ get_inode2 (unsigned int nr) {
 	}
 	return inode;
 }
-#endif
 
 static void
 check_root(void) {
@@ -733,7 +731,6 @@ check_root(void) {
 		die(_("root inode isn't a directory"));
 }
 
-#ifdef HAVE_MINIX2
 static void
 check_root2 (void) {
 	struct minix2_inode *inode = Inode2 + ROOT_INO;
@@ -741,7 +738,6 @@ check_root2 (void) {
 	if (!inode || !S_ISDIR (inode->i_mode))
 		die ("root inode isn't a directory");
 }
-#endif
 
 static int
 add_zone(unsigned short * znr, int * corrected) {
@@ -753,9 +749,9 @@ add_zone(unsigned short * znr, int * corrected) {
 	if (!block)
 		return 0;
 	if (zone_count[block]) {
-		printf(_("Block has been used before. Now in file `"));
-		print_current_name();
-		printf("'.");
+		get_current_name();
+		printf(_("Block has been used before. Now in file `%s'."),
+		       current_name);
 		if (ask(_("Clear"),1)) {
 			*znr = 0;
 			block = 0;
@@ -765,9 +761,9 @@ add_zone(unsigned short * znr, int * corrected) {
 	if (!block)
 		return 0;
 	if (!zone_in_use(block)) {
-		printf(_("Block %d in file `"),block);
-		print_current_name();
-		printf(_("' is marked not in use."));
+		get_current_name();
+		printf(_("Block %d in file `%s' is marked not in use."),
+		       block, current_name);
 		if (ask(_("Correct"),1))
 			mark_zone(block);
 	}
@@ -776,7 +772,6 @@ add_zone(unsigned short * znr, int * corrected) {
 	return block;
 }
 
-#ifdef HAVE_MINIX2
 static int
 add_zone2 (unsigned int *znr, int *corrected) {
 	int result;
@@ -787,9 +782,9 @@ add_zone2 (unsigned int *znr, int *corrected) {
 	if (!block)
 		return 0;
 	if (zone_count[block]) {
-		printf (_("Block has been used before. Now in file `"));
-		print_current_name ();
-		printf ("'.");
+		get_current_name();
+		printf (_("Block has been used before. Now in file `%s'."),
+			current_name);
 		if (ask (_("Clear"), 1)) {
 			*znr = 0;
 			block = 0;
@@ -799,9 +794,9 @@ add_zone2 (unsigned int *znr, int *corrected) {
 	if (!block)
 		return 0;
 	if (!zone_in_use (block)) {
-		printf (_("Block %d in file `"), block);
-		print_current_name ();
-		printf (_("' is marked not in use."));
+		get_current_name();
+		printf (_("Block %d in file `%s' is marked not in use."),
+			block, current_name);
 		if (ask (_("Correct"), 1))
 			mark_zone (block);
 	}
@@ -809,7 +804,6 @@ add_zone2 (unsigned int *znr, int *corrected) {
 		zone_count[block]--;
 	return block;
 }
-#endif
 
 static void
 add_zone_ind(unsigned short * znr, int * corrected) {
@@ -827,7 +821,6 @@ add_zone_ind(unsigned short * znr, int * corrected) {
 		write_block(block, blk);
 }
 
-#ifdef HAVE_MINIX2
 static void
 add_zone_ind2 (unsigned int *znr, int *corrected) {
 	static char blk[BLOCK_SIZE];
@@ -843,7 +836,6 @@ add_zone_ind2 (unsigned int *znr, int *corrected) {
 	if (chg_blk)
 		write_block (block, blk);
 }
-#endif
 
 static void
 add_zone_dind(unsigned short * znr, int * corrected) {
@@ -861,7 +853,6 @@ add_zone_dind(unsigned short * znr, int * corrected) {
 		write_block(block, blk);
 }
 
-#ifdef HAVE_MINIX2
 static void
 add_zone_dind2 (unsigned int *znr, int *corrected) {
 	static char blk[BLOCK_SIZE];
@@ -893,7 +884,6 @@ add_zone_tind2 (unsigned int *znr, int *corrected) {
 	if (blk_chg)
 		write_block (block, blk);
 }
-#endif
 
 static void
 check_zones(unsigned int i) {
@@ -913,7 +903,6 @@ check_zones(unsigned int i) {
 	add_zone_dind(8 + inode->i_zone, &changed);
 }
 
-#ifdef HAVE_MINIX2
 static void
 check_zones2 (unsigned int i) {
 	struct minix2_inode *inode;
@@ -932,7 +921,6 @@ check_zones2 (unsigned int i) {
 	add_zone_dind2 (8 + inode->i_zone, &changed);
 	add_zone_tind2 (9 + inode->i_zone, &changed);
 }
-#endif
 
 static void
 check_file(struct minix_inode * dir, unsigned int offset) {
@@ -947,9 +935,10 @@ check_file(struct minix_inode * dir, unsigned int offset) {
 	name = blk + (offset % BLOCK_SIZE) + 2;
 	ino = * (unsigned short *) (name-2);
 	if (ino > INODES) {
-		print_current_name();
-		printf(_(" contains a bad inode number for file '"));
-		printf("%.*s'.",namelen,name);
+		get_current_name();
+		printf(_("The directory '%s' contains a bad inode number "
+			 "for file '%.*s'."),
+		       current_name, namelen, name);
 		if (ask(_(" Remove"),1)) {
 			*(unsigned short *)(name-2) = 0;
 			write_block(block, blk);
@@ -963,15 +952,17 @@ check_file(struct minix_inode * dir, unsigned int offset) {
 	name_depth--;
 	if (!offset) {
 		if (!inode || strcmp(".",name)) {
-			print_current_name();
-			printf(_(": bad directory: '.' isn't first\n"));
+			get_current_name();
+			printf(_("`%s': bad directory: '.' isn't first\n"),
+			       current_name);
 			errors_uncorrected = 1;
 		} else return;
 	}
 	if (offset == dirsize) {
 		if (!inode || strcmp("..",name)) {
-			print_current_name();
-			printf(_(": bad directory: '..' isn't second\n"));
+			get_current_name();
+			printf(_("`%s': bad directory: '..' isn't second\n"),
+			       current_name);
 			errors_uncorrected = 1;
 		} else return;
 	}
@@ -982,8 +973,10 @@ check_file(struct minix_inode * dir, unsigned int offset) {
 	name_depth++;	
 	if (list) {
 		if (verbose)
-			printf("%6d %07o %3d ",ino,inode->i_mode,inode->i_nlinks);
-		print_current_name();
+			printf("%6d %07o %3d ", ino,
+			       inode->i_mode, inode->i_nlinks);
+		get_current_name();
+		printf("%s", current_name);
 		if (S_ISDIR(inode->i_mode))
 			printf(":\n");
 		else
@@ -996,7 +989,6 @@ check_file(struct minix_inode * dir, unsigned int offset) {
 	return;
 }
 
-#ifdef HAVE_MINIX2
 static void
 check_file2 (struct minix2_inode *dir, unsigned int offset) {
 	static char blk[BLOCK_SIZE];
@@ -1010,9 +1002,10 @@ check_file2 (struct minix2_inode *dir, unsigned int offset) {
 	name = blk + (offset % BLOCK_SIZE) + 2;
 	ino = *(unsigned short *) (name - 2);
 	if (ino > INODES) {
-		print_current_name ();
-		printf (_(" contains a bad inode number for file '"));
-		printf ("%.*s'.", namelen, name);
+		get_current_name();
+		printf(_("The directory '%s' contains a bad inode number "
+			 "for file '%.*s'."),
+			  current_name, namelen, name);
 		if (ask (_(" Remove"), 1)) {
 			*(unsigned short *) (name - 2) = 0;
 			write_block (block, blk);
@@ -1026,16 +1019,18 @@ check_file2 (struct minix2_inode *dir, unsigned int offset) {
 	name_depth--;
 	if (!offset) {
 		if (!inode || strcmp (".", name)) {
-			print_current_name ();
-			printf (_(": bad directory: '.' isn't first\n"));
+			get_current_name ();
+			printf (_("%s: bad directory: '.' isn't first\n"),
+				current_name);
 			errors_uncorrected = 1;
 		} else
 			return;
 	}
 	if (offset == dirsize) {
 		if (!inode || strcmp ("..", name)) {
-			print_current_name ();
-			printf (_(": bad directory: '..' isn't second\n"));
+			get_current_name ();
+			printf (_("%s: bad directory: '..' isn't second\n"),
+				current_name);
 			errors_uncorrected = 1;
 		} else
 			return;
@@ -1045,8 +1040,10 @@ check_file2 (struct minix2_inode *dir, unsigned int offset) {
 	name_depth++;
 	if (list) {
 		if (verbose)
-			printf ("%6d %07o %3d ", ino, inode->i_mode, inode->i_nlinks);
-		print_current_name ();
+			printf ("%6d %07o %3d ", ino, inode->i_mode,
+				inode->i_nlinks);
+		get_current_name ();
+		printf("%s", current_name);
 		if (S_ISDIR (inode->i_mode))
 			printf (":\n");
 		else
@@ -1058,7 +1055,6 @@ check_file2 (struct minix2_inode *dir, unsigned int offset) {
 	name_depth--;
 	return;
 }
-#endif
 
 static void
 recursive_check(unsigned int ino) {
@@ -1069,15 +1065,15 @@ recursive_check(unsigned int ino) {
 	if (!S_ISDIR(dir->i_mode))
 		die(_("internal error"));
 	if (dir->i_size < 2 * dirsize) {
-		print_current_name();
-		printf(_(": bad directory: size<32"));
+		get_current_name();
+		printf(_("%s: bad directory: size < 32"),
+		       current_name);
 		errors_uncorrected = 1;
 	}
 	for (offset = 0 ; offset < dir->i_size ; offset += dirsize)
 		check_file(dir,offset);
 }
 
-#ifdef HAVE_MINIX2
 static void
 recursive_check2 (unsigned int ino) {
 	struct minix2_inode *dir;
@@ -1087,14 +1083,14 @@ recursive_check2 (unsigned int ino) {
 	if (!S_ISDIR (dir->i_mode))
 		die ("internal error");
 	if (dir->i_size < 2 * dirsize) {
-		print_current_name ();
-		printf (_(": bad directory: size < 32"));
+		get_current_name ();
+		printf (_("%s: bad directory: size < 32"),
+			current_name);
 		errors_uncorrected = 1;
 	}
 	for (offset = 0; offset < dir->i_size; offset += dirsize)
 		check_file2 (dir, offset);
 }
-#endif
 
 static int
 bad_zone(int i) {
@@ -1151,12 +1147,15 @@ check_counts(void) {
 				unmark_zone(i);
 			continue;
 		}
-		printf(_("Zone %d: %sin use, counted=%d\n"),
-		i,zone_in_use(i)?"":_("not "),zone_count[i]);
+		if (zone_in_use(i))
+			printf(_("Zone %d: in use, counted=%d\n"),
+			       i, zone_count[i]);
+		else
+			printf(_("Zone %d: not in use, counted=%d\n"),
+			       i, zone_count[i]);
 	}
 }
 
-#ifdef HAVE_MINIX2
 static void
 check_counts2 (void) {
 	int i;
@@ -1197,16 +1196,20 @@ check_counts2 (void) {
 		if (!zone_count[i]) {
 			if (bad_zone (i))
 				continue;
-			printf (_("Zone %d: marked in use, no file uses it."), i);
+			printf (_("Zone %d: marked in use, no file uses it."),
+				i);
 			if (ask (_("Unmark"), 1))
 				unmark_zone (i);
 			continue;
 		}
-		printf (_("Zone %d: %sin use, counted=%d\n"),
-			i, zone_in_use (i) ? "" : _("not "), zone_count[i]);
+		if (zone_in_use (i))
+			printf (_("Zone %d: in use, counted=%d\n"),
+				i, zone_count[i]);
+		else
+			printf (_("Zone %d: not in use, counted=%d\n"),
+				i, zone_count[i]);
 	}
 }
-#endif
 
 static void
 check(void) {
@@ -1217,7 +1220,6 @@ check(void) {
 	check_counts();
 }
 
-#ifdef HAVE_MINIX2
 static void
 check2 (void) {
 	memset (inode_count, 0, (INODES + 1) * sizeof (*inode_count));
@@ -1226,7 +1228,6 @@ check2 (void) {
 	recursive_check2 (ROOT_INO);
 	check_counts2 ();
 }
-#endif
 
 int
 main(int argc, char ** argv) {
@@ -1251,10 +1252,9 @@ main(int argc, char ** argv) {
 
 	if (INODE_SIZE * MINIX_INODES_PER_BLOCK != BLOCK_SIZE)
 		die(_("bad inode size"));
-#ifdef HAVE_MINIX2
 	if (INODE_SIZE2 * MINIX2_INODES_PER_BLOCK != BLOCK_SIZE)
 		die(_("bad v2 inode size"));
-#endif
+
 	while (argc-- > 1) {
 		argv++;
 		if (argv[0][0] != '-') {
@@ -1317,13 +1317,10 @@ main(int argc, char ** argv) {
 		termios_set = 1;
 	}
 
-#if HAVE_MINIX2
 	if (version2) {
 		check_root2 ();
 		check2 ();
-	} else 
-#endif
-	  {
+	} else {
 		check_root();
 		check();
 	}
