@@ -16,7 +16,7 @@
  * Hacked by Peter Breitenlohner, peb@mppmu.mpg.de,
  * to remove trailing empty fields.  Oct 5, 96.
  *
- *  1999-02-22 Arkadiusz Mikiewicz <misiek@misiek.eu.org>
+ *  1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
  *  - added Native Language Support
  *    
  *
@@ -40,6 +40,12 @@
 #include "nls.h"
 #include "env.h"
 
+#ifdef WITH_SELINUX
+#include <selinux/selinux.h>
+#include <selinux/av_permissions.h>
+#include "selinux_utils.h"
+#endif
+
 #if REQUIRE_PASSWORD && USE_PAM
 #include <security/pam_appl.h>
 #include <security/pam_misc.h>
@@ -136,6 +142,27 @@ int main (int argc, char **argv) {
        exit(1);
     }
 
+#ifdef WITH_SELINUX
+    if (is_selinux_enabled()) {
+      if(uid == 0) {
+	if (checkAccess(oldf.username,PASSWD__CHFN)!=0) {
+	  security_context_t user_context;
+	  if (getprevcon(&user_context) < 0)
+	    user_context=(security_context_t) strdup(_("Unknown user context"));
+	  fprintf(stderr, _("%s: %s is not authorized to change the finger info of %s\n"),
+		  whoami, user_context, oldf.username);
+	  freecon(user_context);
+	  exit(1);
+	}
+      }
+      if (setupDefaultContext("/etc/passwd") != 0) {
+	fprintf(stderr,_("%s: Can't set default context for /etc/passwd"),
+		whoami);
+	exit(1);
+      }
+    }
+#endif
+
     /* Reality check */
     if (uid != 0 && uid != oldf.pw->pw_uid) {
 	errno = EACCES;
@@ -157,9 +184,9 @@ int main (int argc, char **argv) {
 	    exit(1);
 	}
         retcode = pam_acct_mgmt(pamh, 0);
-        if (retcode == PAM_NEW_AUTHTOK_REQD) {
+        if (retcode == PAM_NEW_AUTHTOK_REQD)
 	    retcode = pam_chauthtok(pamh, PAM_CHANGE_EXPIRED_AUTHTOK);
-        } else if (retcode) {
+        if (retcode) {
 	    puts(_("Password error."));
 	    exit(1);
 	}
@@ -221,7 +248,7 @@ static boolean parse_argv (argc, argv, pinfo)
     info_given = false;
     while (true) {
 	c = getopt_long (argc, argv, "f:r:p:h:o:uv", long_options, &index);
-	if (c == EOF) break;
+	if (c == -1) break;
 	/* version?  output version and exit. */
 	if (c == 'v') {
 	    printf ("%s\n", util_linux_version);
