@@ -1,4 +1,4 @@
-#!/usr/bin/perl5 -w
+#!/usr/bin/perl -w
 #
 # chkdupexe version 2.1.1
 #
@@ -27,7 +27,10 @@
 $execdirs='/bin /sbin /usr/bin /usr/sbin /usr/local/bin /usr/local/sbin '.
   '/usr/X11/bin /usr/bin/X11 /usr/local/X11/bin '.
   '/usr/TeX/bin /usr/tex/bin /usr/games '.
-  '/usr/local/games /usr/intervies/bin/LINUX';
+  '/usr/local/games';
+
+# Turn off buffering for the output channel.
+$|=1;
 
 # Values from /usr/include/linux/errno.h.  Existence of linux/errno.ph is not
 # something to count on... :-(
@@ -35,7 +38,9 @@ $ENOENT=2;
 
 %didthis=();
 
-foreach $dir (split(/\s+/, "$execdirs")) {
+foreach $dir (split(/\s+/, "$execdirs"), "\0", split(/:/, $ENV{PATH})) {
+
+  if ($dir eq "\0") { $checkingpath = 1; next; }
 
   # It's like this: One directory corresponds to one $device,$inode tuple
   # If a symlink points to a directory we already checked that directory
@@ -52,7 +57,7 @@ foreach $dir (split(/\s+/, "$execdirs")) {
       print "Dangling symlink: $dir\n";
       next;
     }
-    # warn "Nonexistent directory: $dir\n";
+    warn "Nonexistent directory: $dir\n" if ($checkingpath);
     next;
   }
 
@@ -90,12 +95,14 @@ foreach $dir (split(/\s+/, "$execdirs")) {
   closedir(DIR);
 }
 
-open(LS,"| xargs ls -ldU");
+open(LS,"| xargs -r ls -ldU");
 while (($prog,$paths)=each %progs) {
   print LS "$paths\n" if ($count{$prog}>1);
 }
 close(LS);
 
+exit 0;
+
 @unchecked=();
 # Check if the users PATH contains something I've not checked. The site admin
 # might want to know about inconsistencies in user PATHs and chkdupexec 
@@ -108,6 +115,6 @@ foreach $dir (split(/:/,$ENV{'PATH'})) {
   $didthis{$device,$inode}=1;
 }
 
-print "Warning: Your path contanis these directories which chkdupexe have not checked:\n",join(',',@unchecked),
+print "Warning: Your path contains these directories which chkdupexe has not checked:\n",join(',',@unchecked),
   ".\nPlease review the execdirs list in chkdupexe.\n"
     if ($#unchecked>=$[);
