@@ -39,7 +39,7 @@ the maximum number of symbolic links this system can have.
 The program exits with a 1 status ONLY if it finds it cannot
 chdir to /,  or if it encounters an unknown file type.
 
-1999-02-22 Arkadiusz Mikiewicz <misiek@misiek.eu.org>
+1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
 - added Native Language Support
 
 -------------------------------------------------------------*/
@@ -47,6 +47,7 @@ chdir to /,  or if it encounters an unknown file type.
 #include <stdio.h>
 #include <unistd.h>
 #include <string.h>
+#include <strings.h>
 #include <stdlib.h>
 #include <errno.h>
 #include <sys/types.h>
@@ -81,7 +82,7 @@ main(int argc, char **argv) {
     if(argc < 2)
 	usage();
 
-    while((c = getopt(argc, argv, "mx")) != EOF){
+    while((c = getopt(argc, argv, "mx")) != -1){
 	switch(c){
 	    case 'm':
 		mflag = !mflag;
@@ -98,7 +99,9 @@ main(int argc, char **argv) {
     }
 
     if(getcwd(curdir, sizeof(curdir)) == NULL){
-	(void)fprintf(stderr, _("namei: unable to get current directory - %s\n"), curdir);
+	(void)fprintf(stderr,
+		      _("namei: unable to get current directory - %s\n"),
+		      curdir);
 	exit(1);
     }
 
@@ -109,7 +112,9 @@ main(int argc, char **argv) {
 	namei(argv[optind], 0);
 
 	if(chdir(curdir) == -1){
-	    (void)fprintf(stderr, _("namei: unable to chdir to %s - %s (%d)\n"), curdir, ERR);
+	    (void)fprintf(stderr,
+			  _("namei: unable to chdir to %s - %s (%d)\n"),
+			  curdir, ERR);
 	    exit(1);
 	}
     }
@@ -163,6 +168,11 @@ namei(char *file, int lev) {
 
     for(;;){
 
+	if (strlen(file) >= BUFSIZ) {
+		fprintf(stderr,_("namei: buf overflow\n"));
+		return;
+	}
+
 	/*
 	 * Copy up to the next / (or nil) into buf
 	 */
@@ -228,7 +238,7 @@ namei(char *file, int lev) {
 
 	    case S_IFLNK:
 		/*
-		 * Sigh, another symlink.  Read it's contents and
+		 * Sigh, another symlink.  Read its contents and
 		 * call namei()
 		 */
 		
