@@ -16,7 +16,7 @@
  *
  * Converted to terminfo by Kars de Jong (jongk@cs.utwente.nl)
  *
- * 1999-02-22 Arkadiusz Mikiewicz <misiek@misiek.eu.org>
+ * 1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
  * - added Native Language Support
  *
  *
@@ -107,6 +107,7 @@
 #else
 #include <curses.h>
 #endif
+#include <sys/param.h>		/* for MAXPATHLEN */
 #include <sys/ioctl.h>
 #include <sys/time.h>
 #include "nls.h"
@@ -184,7 +185,7 @@ int opt_cl_all;			/* Clear all or rest. */
 int opt_bl_min;			/* Blank screen. */
 int opt_blength_l;
 int opt_bfreq_f;
-int opt_sn_num = 0;		/* Snap screen. */
+int opt_sn_num;			/* Snap screen. */
 int opt_st_attr;
 int opt_rt_len;			/* regular tab length */
 int opt_tb_array[161];		/* Array for tab list */
@@ -290,10 +291,10 @@ par_color(int argc, char **argv, int *option, int *opt_color, int *bad_arg) {
 			*opt_color = DEFAULT;
 		else if (isdigit(argv[0][0]))
 			*opt_color = atoi(argv[0]);
-		else    
+		else 
 			*bad_arg = TRUE;
     
-		if(*opt_color < 0 || *opt_color > 7)
+		if(*opt_color < 0 || *opt_color > 9 || *opt_color == 8)
 			*bad_arg = TRUE;
 	}
 }
@@ -802,7 +803,8 @@ static char *ti_entry(const char *name) {
 
 	char *buf_ptr;
 
-	if ((buf_ptr = tigetstr(name)) == (char *)-1) buf_ptr = NULL;
+	if ((buf_ptr = tigetstr((char *)name)) == (char *)-1)
+		buf_ptr = NULL;
 	return buf_ptr;
 }
 
@@ -1083,7 +1085,7 @@ perform_sequence(int vcterm) {
 			result = klogctl(6, NULL, 0);
 
 		if (result != 0)
-			printf(_("klogctl error: %s\n"), strerror(result));
+			printf(_("klogctl error: %s\n"), strerror(errno));
 	}
 
 	/* -msglevel [0-8] */
@@ -1091,7 +1093,7 @@ perform_sequence(int vcterm) {
 		/* 8 -- Set level of messages printed to console */
 		result = klogctl(8, NULL, opt_msglevel_num);
 		if (result != 0)
-			printf(_("klogctl error: %s\n"), strerror(result));
+			printf(_("klogctl error: %s\n"), strerror(errno));
 	}
 
 	/* -blength [0-2000] */
@@ -1108,7 +1110,6 @@ perform_sequence(int vcterm) {
 
 static void
 screendump(int vcnum, FILE *F) {
-#include <sys/param.h>
     char infile[MAXPATHLEN];
     unsigned char header[4];
     unsigned int rows, cols;
@@ -1116,13 +1117,28 @@ screendump(int vcnum, FILE *F) {
     char *inbuf, *outbuf, *p, *q;
 
     sprintf(infile, "/dev/vcsa%d", vcnum);
-    fd = open(infile, 0);
-    if (fd < 0 || read(fd, header, 4) != 4)
-      goto try_ioctl;
+    fd = open(infile, O_RDONLY);
+    if (fd < 0 && vcnum == 0) {
+	/* vcsa0 is often called vcsa */
+	sprintf(infile, "/dev/vcsa");
+	fd = open(infile, O_RDONLY);
+    }
+    if (fd < 0) {
+	/* try devfs name - for zero vcnum just /dev/vcc/a */
+	/* some gcc's warn for %.u - add 0 */
+	sprintf(infile, "/dev/vcc/a%.0u", vcnum);
+	fd = open(infile, O_RDONLY);
+    }
+    if (fd < 0) {
+	sprintf(infile, "/dev/vcsa%d", vcnum);
+	goto try_ioctl;
+    }
+    if (read(fd, header, 4) != 4)
+	goto try_ioctl;
     rows = header[0];
     cols = header[1];
     if (rows * cols == 0)
-      goto try_ioctl;
+        goto try_ioctl;
     inbuf = malloc(rows*cols*2);
     outbuf = malloc(rows*(cols+1));
     if(!inbuf || !outbuf) {
