@@ -23,6 +23,10 @@
  *               in mtab, try them all, with last one tried first
  *             - Differentiate "user" and "users" key words in fstab
  * 001202: aeb - remove at most one line from /etc/mtab
+ * 010716: Michael K. Johnson <johnsonm@redhat.com: -a -O
+ * 010914: Jamie Strandboge - use tcp if that was used for mount
+ * 011005: hch - add lazy umount support
+ * 020105: aeb - permission test owner umount
  */
 
 #include <stdio.h>
@@ -39,6 +43,7 @@
 #include "lomount.h"
 #include "loop.h"
 #include "fstab.h"
+#include "env.h"
 #include "nls.h"
 
 #ifdef HAVE_NFS
@@ -82,9 +87,16 @@ umount2(const char *path, int flags) {
 
 #endif /* MNT_FORCE */
 
+#if !defined(MNT_DETACH)
+#define MNT_DETACH 2
+#endif
+
 /* Nonzero for force umount (-f).  There is kernel support since 2.1.116.  */
 int force = 0;
 
+/* Nonzero for lazy umount (-l).  There is kernel support since 2.4.11.  */
+int lazy = 0;
+
 /* When umount fails, attempt a read-only remount (-r). */
 int remount = 0;
 
@@ -100,6 +112,16 @@ int verbose = 0;
 /* True if ruid != euid.  */
 int suid = 0;
 
+#ifdef USE_SPECIAL_UMOUNTPROG
+/* unimplemented so far */
+static int
+check_special_umountprog() {
+	/* find type from command line or /etc/mtab;
+	   stat /sbin/umount.%s
+	   if it exists, use it */
+}
+#endif
+
 #ifdef HAVE_NFS
 static int xdr_dir(XDR *xdrsp, char *dirp)
 {
@@ -160,10 +182,20 @@ nfs_umount_rpc_call(const char *spec, const char *opts)
       saddr.sin_port = htons(port);
       pertry.tv_sec = 3;
       pertry.tv_usec = 0;
-      if ((clp = clntudp_create(&saddr, MOUNTPROG, MOUNTVERS,
-				pertry, &so)) == NULL) {
-	   clnt_pcreateerror("Cannot MOUNTPROG RPC");
-	   return (1);
+      if (opts && (p = strstr(opts, "tcp"))) {
+	   /* possibly: make sure option is not "notcp"
+	      possibly: try udp if tcp fails */
+	   if ((clp = clnttcp_create(&saddr, MOUNTPROG, MOUNTVERS,
+				     &so, 0, 0)) == NULL) {
+		clnt_pcreateerror("Cannot MOUNTPROG RPC (tcp)");
+		return 1;
+	   }
+      } else {
+           if ((clp = clntudp_create(&saddr, MOUNTPROG, MOUNTVERS,
+				     pertry, &so)) == NULL) {
+		clnt_pcreateerror("Cannot MOUNTPROG RPC");
+		return 1;
+	   }
       }
       clp->cl_auth = authunix_create_default();
       try.tv_sec = 20;
@@ -175,12 +207,12 @@ nfs_umount_rpc_call(const char *spec, const char *opts)
 
       if (clnt_stat != RPC_SUCCESS) {
 	   clnt_perror(clp, "Bad UMNT RPC");
-	   return (1);
+	   return 1;
       }
       auth_destroy(clp->cl_auth);
       clnt_destroy(clp);
 
-      return (0);
+      return 0;
 }
 #endif /* HAVE_NFS */
 
@@ -236,15 +268,16 @@ umount_one (const char *spec, const char *node, const char *type,
 		nfs_umount_rpc_call(spec, opts);
 #endif
  
-
 	umnt_err = umnt_err2 = 0;
-	if (force) {
-		/* completely untested;
-		   2.1.116 only has some support in nfs case */
-		/* probably this won't work */
-		int flags = MNT_FORCE;
+	if (lazy) {
+		res = umount2 (node, MNT_DETACH);
+		if (res < 0)
+			umnt_err = errno;
+		goto writemtab;
+	}
 
-		res = umount2 (node, flags);
+	if (force) {		/* only supported for NFS */
+		res = umount2 (node, MNT_FORCE);
 		if (res == -1) {
 			perror("umount2");
 			if (errno == ENOSYS) {
@@ -340,6 +373,7 @@ umount_one (const char *spec, const char *node, const char *type,
 	if (loopdev)
 		del_loop(loopdev);
 
+ writemtab:
 	if (!nomtab && mtab_is_writable() &&
 	    (umnt_err == 0 || umnt_err == EINVAL || umnt_err == ENOENT)) {
 		update_mtab (node, NULL);
@@ -366,15 +400,23 @@ umount_one (const char *spec, const char *node, const char *type,
  * In both cases, it is best to try the last occurrence first.
  */
 static int
-umount_one_bw (const char *file, struct mntentchn *mc) {
-     int res = 1;
+umount_one_bw (const char *file, struct mntentchn *mc0) {
+	struct mntentchn *mc;
+	int res = 1;
 
-     while (res && mc) {
-	  res = umount_one(mc->m.mnt_fsname, mc->m.mnt_dir,
-			   mc->m.mnt_type, mc->m.mnt_opts, mc);
-	  mc = getmntfilesbackward (file, mc);
-     }
-     return res;
+	mc = mc0;
+	while (res && mc) {
+		res = umount_one(mc->m.mnt_fsname, mc->m.mnt_dir,
+				 mc->m.mnt_type, mc->m.mnt_opts, mc);
+		mc = getmntdirbackward(file, mc);
+	}
+	mc = mc0;
+	while (res && mc) {
+		res = umount_one(mc->m.mnt_fsname, mc->m.mnt_dir,
+				 mc->m.mnt_type, mc->m.mnt_opts, mc);
+		mc = getmntdevbackward(file, mc);
+	}
+	return res;
 }
 
 /* Unmount all filesystems of type VFSTYPES found in mtab.  Since we are
@@ -383,7 +425,7 @@ umount_one_bw (const char *file, struct mntentchn *mc) {
    in any case it's important to umount mtab entries in reverse order
    to mount, e.g. /usr/spool before /usr.  */
 static int
-umount_all (string_list types) {
+umount_all (char *types, char *test_opts) {
      struct mntentchn *mc, *hd;
      int errors = 0;
 
@@ -391,7 +433,8 @@ umount_all (string_list types) {
      if (!hd->prev)
 	  die (2, _("umount: cannot find list of filesystems to unmount"));
      for (mc = hd->prev; mc != hd; mc = mc->prev) {
-	  if (matching_type (mc->m.mnt_type, types)) {
+	  if (matching_type (mc->m.mnt_type, types)
+	      && matching_opts (mc->m.mnt_opts, test_opts)) {
 	       errors |= umount_one (mc->m.mnt_fsname, mc->m.mnt_dir,
 				     mc->m.mnt_type, mc->m.mnt_opts, mc);
 	  }
@@ -408,6 +451,7 @@ static struct option longopts[] =
   { "force", 0, 0, 'f' },
   { "help", 0, 0, 'h' },
   { "no-mtab", 0, 0, 'n' },
+  { "test-opts", 1, 0, 'O' },
   { "verbose", 0, 0, 'v' },
   { "version", 0, 0, 'V' },
   { "read-only", 0, 0, 'r' },
@@ -419,159 +463,236 @@ static void
 usage (FILE *fp, int n)
 {
   fprintf (fp, _("Usage: umount [-hV]\n"
-	       "       umount -a [-f] [-r] [-n] [-v] [-t vfstypes]\n"
+	       "       umount -a [-f] [-r] [-n] [-v] [-t vfstypes] [-O opts]\n"
 	       "       umount [-f] [-r] [-n] [-v] special | node...\n"));
   exit (n);
 }
 
 int mount_quiet = 0;
 
-int
-main (int argc, char *argv[])
-{
-  int c;
-  int all = 0;
-  string_list types = NULL;
-  string_list options;
-  struct mntentchn *mc, *fs;
-  char *file;
-  int result = 0;
+/*=======================================================================*/
+/* string list stuff - no longer used by mount - will disappear entirely */
+typedef struct string_list {
+	char *hd;
+	struct string_list *tl;
+} *string_list;
 
-  setlocale(LC_ALL, "");
-  bindtextdomain(PACKAGE, LOCALEDIR);
-  textdomain(PACKAGE);
+#define car(p) ((p) -> hd)
+#define cdr(p) ((p) -> tl)
 
-  while ((c = getopt_long (argc, argv, "afhnrt:vV",
-			   longopts, NULL)) != EOF)
-    switch (c) {
-      case 'a':			/* umount everything */
-	++all;
-	break;
-	/* fall through? */
-      case 'd':			/* do losetup -d for unmounted loop devices */
-	++delloop;
-	break;
-      case 'f':			/* force umount */
-	++force;
-	break;
-      case 'h':			/* help */
-	usage (stdout, 0);
-	break;
-      case 'n':			/* do not write in /etc/mtab */
-	++nomtab;
-	break;
-      case 'r':			/* remount read-only if umount fails */
-	++remount;
-	break;
-      case 'v':			/* make noise */
-	++verbose;
-	break;
-      case 'V':			/* version */
-	printf ("umount: %s\n", version);
-	exit (0);
-      case 't':			/* specify file system type */
-	types = parse_list (optarg);
-	break;
-      case 0:
-	break;
-      case '?':
-      default:
-	usage (stderr, 1);
-    }
+static string_list
+cons (char *a, const string_list b) {
+	string_list p;
 
-  if (getuid () != geteuid ())
-    {
-      suid = 1;
-      if (all || types || nomtab || force)
-	die (2, _("umount: only root can do that"));
-    }
-
-  argc -= optind;
-  argv += optind;
-
-  if (all) {
-       if (types == NULL)
-	  types = parse_list("noproc");
-       result = umount_all (types);
-  } else if (argc < 1) {
-       usage (stderr, 2);
-  } else while (argc--) {
-       file = canonicalize (*argv); /* mtab paths are canonicalized */
-       if (verbose > 1)
-	  printf(_("Trying to umount %s\n"), file);
-
-       mc = getmntfilesbackward (file, NULL);
-       if (!mc && verbose)
-	  printf(_("Could not find %s in mtab\n"), file);
-
-       if (suid) {
-	  if (!mc)
-	    die (2, _("umount: %s is not mounted (according to mtab)"), file);
-	  if (getmntfilesbackward (file, mc))
-	    die (2, _("umount: it seems %s is mounted multiple times"), file);
-
-	  /* If fstab contains the two lines
-	       /dev/sda1 /mnt/zip auto user,noauto  0 0
-	       /dev/sda4 /mnt/zip auto user,noauto  0 0
-	     then "mount /dev/sda4" followed by "umount /mnt/zip"
-	     used to fail. So, we must not look for file, but for
-	     the pair (spec,file) in fstab. */
-	  fs = getfsspecfile(mc->m.mnt_fsname, mc->m.mnt_dir);
-	  if (!fs) {
-	    if (!getfsspec (file) && !getfsfile (file))
-	      die (2,
-		 _("umount: %s is not in the fstab (and you are not root)"),
-		 file);
-	    else
-	      die (2, _("umount: %s mount disagrees with the fstab"), file);
-	  }
-
-	  /* User mounting and unmounting is allowed only
-	     if fstab contains the option `user' or `users' */
-	  /* The option `users' allows arbitrary users to mount
-	     and unmount - this may be a security risk. */
-	  /* The option `user' only allows unmounting by the user
-	     that mounted. */
-	  /* The option `owner' only allows (un)mounting by the owner. */
-	  /* A convenient side effect is that the user who mounted
-	     is visible in mtab. */
-	  options = parse_list (fs->m.mnt_opts);
-	  while (options) {
-	      if (streq (car (options), "user") ||
-		  streq (car (options), "users") ||
-		  streq (car (options), "owner"))
-		break;
-	      options = cdr (options);
-	  }
-	  if (!options)
-	    die (2, _("umount: only root can unmount %s from %s"),
-		 fs->m.mnt_fsname, fs->m.mnt_dir);
-	  if (streq (car (options), "user") ||
-	      streq (car (options), "owner")) {
-	      char *user = getusername();
-
-	      options = parse_list (mc->m.mnt_opts);
-	      while (options) {
-		  char *co = car (options);
-		  if (!strncmp(co, "user=", 5)) {
-		      if (!user || !streq(co+5,user))
-			  die(2, _("umount: only %s can unmount %s from %s"),
-			      co+5, fs->m.mnt_fsname, fs->m.mnt_dir);
-		      break;
-		  }
-		  options = cdr (options);
-	      }
-	  }
-       }
-
-       if (mc)
-	    result = umount_one_bw (file, mc);
-       else
-	    result = umount_one (*argv, *argv, *argv, *argv, NULL);
-
-       argv++;
-
-  }
-  exit (result);
+	p = xmalloc (sizeof *p);
+	car (p) = a;
+	cdr (p) = b;
+	return p;
 }
 
+/* Parse a list of strings like str[,str]... into a string list.  */
+static string_list
+parse_list (char *strings) {
+	string_list list;
+	char *s, *t;
+
+	if (strings == NULL)
+		return NULL;
+
+	/* strtok() destroys its argument, so we have to use a copy */
+	s = xstrdup(strings);
+
+	list = cons (strtok (s, ","), NULL);
+
+	while ((t = strtok (NULL, ",")) != NULL)
+		list = cons (t, list);
+
+	return list;
+}
+
+static int
+contains(string_list list, char *s) {
+	while (list) {
+		if (streq (car (list), s))
+			return 1;
+		list = cdr (list);
+	}
+	return 0;
+}
+
+/*
+ * If list contains "user=peter" and we ask for "user=", return "peter"
+ */
+static char *
+get_value(string_list list, char *s) {
+	int n = strlen(s);
+	while (list) {
+		if (strncmp (car (list), s, n) == 0)
+			return car(list)+n;
+		list = cdr (list);
+	}
+	return 0;
+}
+
+static int
+umount_file (char *arg) {
+	struct mntentchn *mc, *fs;
+	char *file;
+	string_list options;
+	int fstab_has_user, fstab_has_users, fstab_has_owner, ok;
+
+	file = canonicalize(arg); /* mtab paths are canonicalized */
+	if (verbose > 1)
+		printf(_("Trying to umount %s\n"), file);
+
+	mc = getmntdirbackward(file, NULL);
+	if (!mc)
+		mc = getmntdevbackward(file, NULL);
+	if (!mc && verbose)
+		printf(_("Could not find %s in mtab\n"), file);
+
+	if (suid) {
+		if (!mc)
+			die(2,
+			    _("umount: %s is not mounted (according to mtab)"),
+			    file);
+		if (!is_mounted_once(file))
+			die(2,
+			    _("umount: it seems %s is mounted multiple times"),
+			    file);
+
+		/* If fstab contains the two lines
+		   /dev/sda1 /mnt/zip auto user,noauto  0 0
+		   /dev/sda4 /mnt/zip auto user,noauto  0 0
+		   then "mount /dev/sda4" followed by "umount /mnt/zip"
+		   used to fail. So, we must not look for file, but for
+		   the pair (spec,file) in fstab. */
+		fs = getfsspecfile(mc->m.mnt_fsname, mc->m.mnt_dir);
+		if (!fs) {
+			if (!getfsspec (file) && !getfsfile (file))
+				die (2,
+				     _("umount: %s is not in the fstab (and you are not root)"),
+				     file);
+			else
+				die (2, _("umount: %s mount disagrees with the fstab"), file);
+		}
+
+		/* User mounting and unmounting is allowed only
+		   if fstab contains one of the options `user',
+		   `users' or `owner'. */
+		/* The option `users' allows arbitrary users to mount
+		   and unmount - this may be a security risk. */
+		/* The option `user' only allows unmounting by the user
+		   that mounted. */
+		/* The option `owner' only allows (un)mounting by the owner. */
+		/* A convenient side effect is that the user who mounted
+		   is visible in mtab. */
+
+		options = parse_list (fs->m.mnt_opts);
+		fstab_has_user = contains(options, "user");
+		fstab_has_users = contains(options, "users");
+		fstab_has_owner = contains(options, "owner");
+		ok = 0;
+
+		if (fstab_has_users)
+			ok = 1;
+
+		if (!ok && (fstab_has_user || fstab_has_owner)) {
+			char *user = getusername();
+			char *mtab_user;
+
+			options = parse_list (mc->m.mnt_opts);
+			mtab_user = get_value(options, "user=");
+
+			if (user && mtab_user && streq (user, mtab_user))
+				ok = 1;
+		}
+		if (!ok)
+			die (2, _("umount: only root can unmount %s from %s"),
+			     fs->m.mnt_fsname, fs->m.mnt_dir);
+	}
+
+	if (mc)
+		return umount_one_bw (file, mc);
+	else
+		return umount_one (arg, arg, arg, arg, NULL);
+}
+
+int
+main (int argc, char *argv[]) {
+	int c;
+	int all = 0;
+	char *types = NULL, *test_opts = NULL;
+	int result = 0;
+
+	sanitize_env();
+	setlocale(LC_ALL, "");
+	bindtextdomain(PACKAGE, LOCALEDIR);
+	textdomain(PACKAGE);
+
+	umask(033);
+
+	while ((c = getopt_long (argc, argv, "adfhlnrt:O:vV",
+				 longopts, NULL)) != -1)
+		switch (c) {
+		case 'a':		/* umount everything */
+			++all;
+			break;
+			/* fall through? */
+		case 'd':		/* do losetup -d for unmounted loop devices */
+			++delloop;
+			break;
+		case 'f':		/* force umount */
+			++force;
+			break;
+		case 'h':		/* help */
+			usage (stdout, 0);
+			break;
+		case 'l':		/* lazy umount */
+			++lazy;
+			break;
+		case 'n':		/* do not write in /etc/mtab */
+			++nomtab;
+			break;
+		case 'O':		/* specify file system options */
+			test_opts = optarg;
+			break;
+		case 'r':		/* remount read-only if umount fails */
+			++remount;
+			break;
+		case 'v':		/* make noise */
+			++verbose;
+			break;
+		case 'V':		/* version */
+			printf ("umount: %s\n", version);
+			exit (0);
+		case 't':		/* specify file system type */
+			types = optarg;
+			break;
+		case 0:
+			break;
+		case '?':
+		default:
+			usage (stderr, 1);
+		}
+
+	if (getuid () != geteuid ()) {
+		suid = 1;
+		if (all || types || nomtab || force)
+			die (2, _("umount: only root can do that"));
+	}
+
+	argc -= optind;
+	argv += optind;
+
+	if (all) {
+		if (types == NULL)
+			types = "noproc,nodevfs";
+		result = umount_all (types, test_opts);
+	} else if (argc < 1) {
+		usage (stderr, 2);
+	} else while (argc--) {
+		result += umount_file(*argv++);
+	}
+	exit (result);		/* nonzero on at least one failure */
+}
