@@ -26,7 +26,7 @@
 **	  on linux/axp.
 **	modified by Kars de Jong <jongk@cs.utwente.nl> to use terminfo instead
 **	  of termcap.
-	1999-02-22 Arkadiusz Mikiewicz <misiek@misiek.eu.org>
+	1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
 	- added Native Language Support
 	1999-03-19 Arnaldo Carvalho de Melo <acme@conectiva.com.br>
 	- more nls translatable strings
@@ -57,12 +57,12 @@
 #include <locale.h>
 #include "xstrncpy.h"
 #include "nls.h"
+#include "widechar.h"
 
 #define _REGEX_RE_COMP
 #include <regex.h>
 #undef _REGEX_RE_COMP
 
-/* #define MOREHELPFILE "/usr/lib/more.help" */
 #define VI		"vi"	/* found on the user's path */
 
 #define Fopen(s,m)	(Currline = 0,file_pos=0,fopen(s,m))
@@ -70,6 +70,9 @@
 #define Fseek(f,off)	(file_pos=off,fseek(f,off,0))
 #define Getc(f)		(++file_pos, getc(f))
 #define Ungetc(c,f)	(--file_pos, ungetc(c,f))
+#define putcerr(c)	fputc(c, stderr)
+#define putserr(s)	fputs(s, stderr)
+#define putsout(s)	fputs(s, stdout)
 
 #define stty(fd,argp)  tcsetattr(fd,TCSANOW,argp)
 
@@ -94,19 +97,14 @@ void screen (register FILE *f, register int num_lines);
 int  command (char *filename, register FILE *f);
 void erasep (register int col);
 void show (register char ch);
-int  pr(char *s1);
-int  printd (int n);
 void set_tty(void);
 void reset_tty(void);
-void ttyin (char buf[], register int nmax, char pchar);
+void ttyin (unsigned char buf[], register int nmax, char pchar);
 int  number(char *cmd);
 int  readch (void);
 int  get_line(register FILE *f, int *length);
 void prbuf (register char *s, register int n);
-int  xprintf (char *fmt, ...);
 void execute (char *filename, char *cmd, ...);
-void errwrite (char *txt);
-void errwrite1 (char *sym);
 FILE *checkf (char *, int *);
 
 #define TBUFSIZ	1024
@@ -182,7 +180,7 @@ extern char	PC;		/* pad character */
 
 static void
 my_putstring(char *s) {
-     putp(s);
+	tputs (s, 1, putchar);		/* putp(s); */
 }
 
 static void
@@ -320,8 +318,7 @@ int main(int argc, char **argv) {
     /* allow clreol only if Home and eraseln and EodClr strings are
      *  defined, and in that case, make sure we are in noscroll mode
      */
-    if(clreol)
-    {
+    if (clreol) {
         if((Home == NULL) || (*Home == '\0') ||
 	   (eraseln == NULL) || (*eraseln == '\0') ||
            (EodClr == NULL) || (*EodClr == '\0') )
@@ -329,7 +326,7 @@ int main(int argc, char **argv) {
 	else noscroll = 1;
     }
     if (dlines == 0)
-	dlines = Lpp - (noscroll ? 1 : 2);
+	    dlines = Lpp - 1;	/* was: Lpp - (noscroll ? 1 : 2) */
     left = dlines;
     if (nfiles > 1)
 	prnames++;
@@ -412,14 +409,14 @@ int main(int argc, char **argv) {
 			erasep (0);
 		    if (clreol)
 			cleareol ();
-		    pr("::::::::::::::");
+		    putsout("::::::::::::::");
 		    if (promptlen > 14)
 			erasep (14);
-		    xprintf ("\n");
+		    putchar('\n');
 		    if(clreol) cleareol();
-		    xprintf("%s\n", fnames[fnum]);
+		    puts(fnames[fnum]);
 		    if(clreol) cleareol();
-		    xprintf("::::::::::::::\n");
+		    puts("::::::::::::::");
 		    if (left > Lpp - 4)
 			left = Lpp - 4;
 		}
@@ -483,10 +480,8 @@ void argscan(char *s, char *argv0) {
 		  case '-': case ' ': case '\t':
 			break;
 		  default:
-			fputs(argv0,stderr);
-			fputs(": unknown option \"-",stderr);
-			putc(*s,stderr);
-			fputs("\"\n",stderr);
+			fprintf(stderr,
+				_("%s: unknown option \"-%c\"\n"), argv0, *s);
 			usage(argv0);
 			exit(1);
 			break;
@@ -508,7 +503,7 @@ checkf (fs, clearfirst)
 {
 	struct stat stbuf;
 	register FILE *f;
-	char c;
+	int c;
 
 	if (stat (fs, &stbuf) == -1) {
 		(void)fflush(stdout);
@@ -518,7 +513,7 @@ checkf (fs, clearfirst)
 		return((FILE *)NULL);
 	}
 	if ((stbuf.st_mode & S_IFMT) == S_IFDIR) {
-		xprintf(_("\n*** %s: directory ***\n\n"), fs);
+		printf(_("\n*** %s: directory ***\n\n"), fs);
 		return((FILE *)NULL);
 	}
 	if ((f = Fopen(fs, "r")) == NULL) {
@@ -529,7 +524,7 @@ checkf (fs, clearfirst)
 	if (magic(f, fs))
 		return((FILE *)NULL);
 	c = Getc(f);
-	*clearfirst = c == '\f';
+	*clearfirst = (c == '\f');
 	Ungetc (c, f);
 	if ((file_size = stbuf.st_size) == 0)
 		file_size = LONG_MAX;
@@ -548,6 +543,10 @@ magic(f, fs)
 {
 	char twobytes[2];
 
+	/* don't try to look ahead if the input is unseekable */
+	if (fseek(f, 0L, SEEK_SET))
+		return(0);
+
 	if (fread(twobytes, 2, 1, f) == 1) {
 		switch(twobytes[0] + (twobytes[1]<<8)) {
 		case OMAGIC:	/* 0407 */
@@ -557,7 +556,7 @@ magic(f, fs)
 		case 0411:
 		case 0177545:
 		case 0x457f:		/* simple ELF detection */
-			xprintf(_("\n******** %s: Not a text file ********\n\n"), fs);
+			printf(_("\n******** %s: Not a text file ********\n\n"), fs);
 			(void)fclose(f);
 			return(1);
 		}
@@ -660,9 +659,7 @@ void onquit(int dummy) {
 	    Pause++;
     }
     else if (!dum_opt && notell) {
-	char *s = _("[Use q or Q to quit]");
-	errwrite(s);
-	promptlen += strlen(s);
+	promptlen += fprintf(stderr, _("[Use q or Q to quit]"));
 	notell = 0;
     }
     signal(SIGQUIT, onquit);
@@ -683,7 +680,7 @@ void chgwinsz(int dummy) {
 	    nscroll = Lpp/2 - 1;
 	    if (nscroll <= 0)
 		nscroll = 1;
-	    dlines = Lpp - (noscroll ? 1 : 2);
+	    dlines = Lpp - 1;	/* was: Lpp - (noscroll ? 1 : 2) */
 	}
 	if (win.ws_col != 0)
 	    Mcol = win.ws_col;
@@ -708,7 +705,7 @@ void end_it (int dummy) {
 	fflush (stdout);
     }
     else
-	errwrite("\n");
+	putcerr('\n');
     _exit(0);
 }
 
@@ -719,93 +716,7 @@ void copy_file(register FILE *f) {
 	putchar(c);
 }
 
-/* Simplified printf function */
-
-int xprintf (char *fmt, ...) {
-	va_list ap;
-	char ch;
-	int ccount;
-
-	ccount = 0;
-	va_start(ap, fmt);
-	while (*fmt) {
-		while ((ch = *fmt++) != '%') {
-			if (ch == '\0')
-				return (ccount);
-			ccount++;
-			putchar (ch);
-		}
-		switch (*fmt++) {
-		case 'd':
-			ccount += printd (va_arg(ap, int));
-			break;
-		case 's':
-			ccount += pr (va_arg(ap, char *));
-			break;
-		case '%':
-			ccount++;
-			putchar ('%');
-			break;
-		case '0':
-			return (ccount);
-		default:
-			break;
-		}
-	}
-	va_end(ap);
-	return (ccount);
-
-}
-
-/*
-** Print an integer as a string of decimal digits,
-** returning the length of the print representation.
-*/
-
-int printd (int n)
-{
-    int a, nchars;
-
-    if ((a = n/10) != 0)
-	nchars = 1 + printd(a);
-    else
-	nchars = 1;
-    putchar (n % 10 + '0');
-    return (nchars);
-}
-
-/* Put the print representation of an integer into a string */
-static char *sptr;
-
-static void Sprintf (int n) {
-    int a;
-
-    if ((a = n/10) != 0)
-	Sprintf (a);
-    *sptr++ = n % 10 + '0';
-}
-
-static void scanstr (int n, char *str)
-{
-    sptr = str;
-    Sprintf (n);
-    *sptr = '\0';
-}
-
-#define ringbell()	errwrite("\007");
-
-#ifdef undef
-strlen (s)
-char *s;
-{
-    register char *p;
-
-    p = s;
-    while (*p++)
-	;
-    return (p - s - 1);
-}
-#endif
+#define ringbell()	putcerr('\007')
 
 /* See whether the last component of the path name "path" is equal to the
 ** string "string"
@@ -840,15 +751,14 @@ static void prompt (char *filename)
 	}
 	if (clreol)
 	    cleareol ();
-	promptlen += pr(_("--More--"));
+	promptlen += printf(_("--More--"));
 	if (filename != NULL) {
-	    promptlen += xprintf (_("(Next file: %s)"), filename);
+	    promptlen += printf(_("(Next file: %s)"), filename);
 	} else if (!no_intty) {
-	    promptlen += xprintf ("(%d%%)",
-				  (int)((file_pos * 100) / file_size));
+	    promptlen += printf("(%d%%)", (int) ((file_pos * 100) / file_size));
 	}
 	if (dum_opt) {
-	    promptlen += pr(_("[Press space to continue, 'q' to quit.]"));
+	    promptlen += printf(_("[Press space to continue, 'q' to quit.]"));
 	}
 	if (Senter && Sexit)
 	    my_putstring (Sexit);
@@ -862,15 +772,30 @@ static void prompt (char *filename)
 }
 
 /*
-** Get a logical line
-*/
+ * Get a logical line
+ */
 
 int get_line(register FILE *f, int *length)
 {
-    register int	c;
-    register char	*p;
-    register int	column;
-    static int		colflg;
+    int	c;
+    char *p;
+    int	column;
+    static int colflg;
+
+#ifdef ENABLE_WIDECHAR
+    int i;
+    wchar_t wc;
+    int wc_width;
+    mbstate_t state, state_bak;		/* Current status of the stream. */
+    unsigned char mbc[MB_LEN_MAX];	/* Buffer for one multibyte char. */
+    size_t mblength;			/* Byte length of multibyte char. */
+    size_t mbc_pos = 0;			/* Postion of the MBC. */
+    int use_mbc_buffer_flag = 0;	/* If 1, mbc has data. */
+    int break_flag = 0;			/* If 1, exit while(). */
+    long file_pos_bak = Ftell (f);
+
+    memset (&state, '\0', sizeof (mbstate_t));
+#endif
 
     p = Line;
     column = 0;
@@ -880,6 +805,58 @@ int get_line(register FILE *f, int *length)
 	c = Getc (f);
     }
     while (p < &Line[LINSIZ - 1]) {
+#ifdef ENABLE_WIDECHAR
+	if (fold_opt && use_mbc_buffer_flag && MB_CUR_MAX > 1) {
+	    use_mbc_buffer_flag = 0;
+	    state_bak = state;
+	    mbc[mbc_pos++] = c;
+process_mbc:
+	    mblength = mbrtowc (&wc, mbc, mbc_pos, &state);
+
+	    switch (mblength) {
+	      case (size_t)-2:	  /* Incomplete multibyte character. */
+		use_mbc_buffer_flag = 1;
+		state = state_bak;
+		break;
+
+	      case (size_t)-1:	  /* Invalid as a multibyte character. */
+		*p++ = mbc[0];
+		state = state_bak;
+		column++;
+		file_pos_bak++;
+		
+		if (column >= Mcol) {
+		    Fseek (f, file_pos_bak);
+		} else {
+		    memmove (mbc, mbc + 1, --mbc_pos);
+		    if (mbc_pos > 0) {
+			mbc[mbc_pos] = '\0';
+			goto process_mbc;
+		    }
+		}
+		break;
+
+	      default:
+		wc_width = wcwidth (wc);
+
+		if (column + wc_width > Mcol) {
+		    Fseek (f, file_pos_bak);
+		    break_flag = 1;
+		} else {
+		    for (i = 0; i < mbc_pos; i++)
+		      *p++ = mbc[i];
+		    if (wc_width > 0)
+		      column += wc_width;
+		}
+	    }
+
+	    if (break_flag || column >= Mcol)
+	      break;
+
+	    c = Getc (f);
+	    continue;
+	}
+#endif 
 	if (c == EOF) {
 	    if (p > Line) {
 		*p = '\0';
@@ -893,6 +870,7 @@ int get_line(register FILE *f, int *length)
 	    Currline++;
 	    break;
 	}
+
 	*p++ = c;
 #if 0
 	if (c == '\033') {      /* ESC */
@@ -923,8 +901,7 @@ int get_line(register FILE *f, int *length)
 		    }
 		    if (column >= promptlen) promptlen = 0;
 		}
-	    }
-	    else
+	    } else
 		column = 1 + (column | 7);
 	} else if (c == '\b' && column > 0) {
 	    column--;
@@ -945,30 +922,43 @@ int get_line(register FILE *f, int *length)
 	} else if (c == EOF) {
 	    *length = p - Line;
 	    return (column);
-	}
-#if 0
-	else {
-		char ch[2]; /* for mblen() */
+	} else {
+#ifdef ENABLE_WIDECHAR
+	    if (fold_opt && MB_CUR_MAX > 1) {
+		memset (mbc, '\0', MB_LEN_MAX);
+		mbc_pos = 0;
+		mbc[mbc_pos++] = c;
+		state_bak = state;
 
-		ch[0] = c;
-		ch[1] = '\0';
-		if (mblen(ch, 1) <= 0) {    /* multibyte, maybe 2 columns */
-			if (column == Mcol-1 && fold_opt) {
-				p--;
-				Ungetc(c, f);
-				break;
-			}
-			*p++ = Getc(f);
-			column += 2;
-		} else if (c & 0x80) {      /* maybe SJIS KANA, or utf-8 ? */
-			column++;
-		} else if (isprint(c))
-			column++;
-	}
-#else
-	else if (isprint(c))
-		column++;
+		mblength = mbrtowc (&wc, mbc, mbc_pos, &state);
+
+		/* The value of mblength is always less than 2 here. */
+		switch (mblength) {
+		  case (size_t)-2:
+		    p--;
+		    file_pos_bak = Ftell (f) - 1;
+		    state = state_bak;
+		    use_mbc_buffer_flag = 1;
+		    break;
+
+		  case (size_t)-1:
+		    state = state_bak;
+		    column++;
+		    break;
+
+		  default:
+		    wc_width = wcwidth (wc);
+		    if (wc_width > 0)
+		      column += wc_width;
+		}
+	    } else
 #endif
+	      {
+		if (isprint(c))
+		   column++;
+	      }
+	}
+
 	if (column >= Mcol && fold_opt)
 		break;
 	c = Getc (f);
@@ -1015,11 +1005,11 @@ void erasep (register int col)
 ** Erase the current line entirely
 */
 
-void kill_line ()
+void kill_line()
 {
-    erasep (0);
+    erasep(0);
     if (!eraseln || dumb)
-	putchar ('\r');
+	putchar('\r');
 }
 
 /*
@@ -1035,21 +1025,6 @@ void clreos()
     my_putstring(EodClr);
 }
 
-/*
-**  Print string and return number of characters
-*/
-
-int pr(char *s1)
-{
-    register char	*s;
-    register char	c;
-
-    for (s = s1; (c = *s++) != 0; )
-	putchar(c);
-    return (int) (s - s1 - 1);
-}
-
-
 /* Print a buffer of n characters */
 
 void prbuf (register char *s, register int n)
@@ -1081,7 +1056,7 @@ void prbuf (register char *s, register int n)
 	    if (c != ' ' || pstate == 0 || state != 0 || ulglitch == 0)
 	        putchar(c);
 	    if (state && *chUL) {
-		pr(chBS);
+		putsout(chBS);
 		my_putstring(chUL);
 	    }
 	    pstate = state;
@@ -1131,7 +1106,6 @@ int command (char *filename, register FILE *f)
     register int retval = 0;
     register int c;
     char colonch;
-    FILE *helpf;
     int done;
     char comchar, cmdbuf[80];
 
@@ -1179,18 +1153,16 @@ int command (char *filename, register FILE *f)
 
 		putchar ('\r');
 		erasep (0);
-		xprintf ("\n");
+		putchar('\n');
 		if (clreol)
 			cleareol ();
-		xprintf (_("...back %d page"), nlines);
-		if (nlines > 1)
-			pr ("s\n");
+		if (nlines != 1)
+			printf(_("...back %d pages"), nlines);
 		else
-			pr ("\n");
-
+			putsout(_("...back 1 page"));
 		if (clreol)
 			cleareol ();
-		pr ("\n");
+		putchar('\n');
 
 		initline = Currline - dlines * (nlines + 1);
 		if (! noscroll)
@@ -1220,23 +1192,23 @@ int command (char *filename, register FILE *f)
 	    end_it (0);
 	case 's':
 	case 'f':
+	case ctrl('F'):
 	    if (nlines == 0) nlines++;
 	    if (comchar == 'f')
 		nlines *= dlines;
 	    putchar ('\r');
 	    erasep (0);
-	    xprintf ("\n");
+	    putchar('\n');
 	    if (clreol)
 		cleareol ();
-	    xprintf (_("...skipping %d line"), nlines);
-	    if (nlines > 1)
-		pr ("s\n");
+	    if (nlines == 1)
+		    putsout(_("...skipping one line"));
 	    else
-		pr ("\n");
+		    printf(_("...skipping %d lines"), nlines);
 
 	    if (clreol)
 		cleareol ();
-	    pr ("\n");
+	    putchar('\n');
 
 	    while (nlines > 0) {
 		while ((c = Getc (f)) != '\n')
@@ -1269,7 +1241,7 @@ int command (char *filename, register FILE *f)
 	case '\'':
 	    if (!no_intty) {
 		kill_line ();
-		pr (_("\n***Back***\n\n"));
+		putsout(_("\n***Back***\n\n"));
 		Fseek (f, context.chrctr);
 		Currline = context.line;
 		ret (dlines);
@@ -1280,7 +1252,7 @@ int command (char *filename, register FILE *f)
 	    }
 	case '=':
 	    kill_line ();
-	    promptlen = printd (Currline);
+	    promptlen = printf("%d", Currline);
 	    fflush (stdout);
 	    break;
 	case 'n':
@@ -1288,16 +1260,16 @@ int command (char *filename, register FILE *f)
 	case '/':
 	    if (nlines == 0) nlines++;
 	    kill_line ();
-	    pr ("/");
+	    putchar('/');
 	    promptlen = 1;
 	    fflush (stdout);
 	    if (lastp) {
-		errwrite ("\r");
+		putcerr('\r');
 		search (NULL, f, nlines);	/* Use previous r.e. */
 	    }
 	    else {
 		ttyin (cmdbuf, sizeof(cmdbuf)-2, '/');
-		errwrite("\r");
+		putcerr('\r');
 		search (cmdbuf, f, nlines);
 	    }
 	    ret (dlines-1);
@@ -1306,35 +1278,47 @@ int command (char *filename, register FILE *f)
 	    break;
 	case '?':
 	case 'h':
-	    helpf = NULL;
-	    {
-		    char *lang;
-		    char hlpfile[sizeof(MOREHELPFILE)+4];
-
-		    lang = getenv("LANGUAGE");
-		    if (!lang || *lang == '\0')
-			    lang = setlocale(LC_MESSAGES, "");
-		    if (!lang || *lang == '\0')
-			    lang = getenv("LANG");
-		    if (lang && strlen(lang) > 1) {
-			    strcpy(hlpfile, MOREHELPFILE);
-			    strcat(hlpfile, ".");
-			    strncat(hlpfile, lang, 2);
-			    helpf = fopen (hlpfile, "r");
-		    }
-	    }
-	    if (helpf == NULL)
-		    helpf = fopen (MOREHELPFILE, "r");
-	    if (helpf == NULL)
-		error (_("Can't open help file"));
-	    if (noscroll) doclear ();
-	    copy_file (helpf);
-	    fclose (helpf);
-	    prompt (filename);
+	    if (noscroll) doclear();
+	    putsout(_("\n"
+"Most commands optionally preceded by integer argument k.  "
+"Defaults in brackets.\n"
+"Star (*) indicates argument becomes new default.\n"));
+	    puts("---------------------------------------"
+		"----------------------------------------");
+	    putsout(_(
+"<space>                 Display next k lines of text [current screen size]\n"
+"z                       Display next k lines of text [current screen size]*\n"
+"<return>                Display next k lines of text [1]*\n"
+"d or ctrl-D             Scroll k lines [current scroll size, initially 11]*\n"
+"q or Q or <interrupt>   Exit from more\n"
+"s                       Skip forward k lines of text [1]\n"
+"f                       Skip forward k screenfuls of text [1]\n"
+"b or ctrl-B             Skip backwards k screenfuls of text [1]\n"
+"'                       Go to place where previous search started\n"
+"=                       Display current line number\n"
+"/<regular expression>   Search for kth occurrence of regular expression [1]\n"
+"n                       Search for kth occurrence of last r.e [1]\n"
+"!<cmd> or :!<cmd>       Execute <cmd> in a subshell\n"
+"v                       Start up /usr/bin/vi at current line\n"
+"ctrl-L                  Redraw screen\n"
+":n                      Go to kth next file [1]\n"
+":p                      Go to kth previous file [1]\n"
+":f                      Display current file name and line number\n"
+".                       Repeat previous command\n"));
+	    puts("---------------------------------------"
+		"----------------------------------------");
+	    prompt(filename);
 	    break;
 	case 'v':	/* This case should go right before default */
 	    if (!no_intty) {
-		    char *editor;
+		    /*
+		     * Earlier: call vi +n file. This also works for emacs.
+		     * POSIX: call vi -c n file (when editor is vi or ex).
+		     */
+		    char *editor, *p;
+		    int n = (Currline - dlines <= 0 ? 1 :
+			     Currline - (dlines + 1) / 2);
+		    int split = 0;
 
 		    editor = getenv("VISUAL");
 		    if (editor == NULL || *editor == '\0')
@@ -1342,25 +1326,41 @@ int command (char *filename, register FILE *f)
 		    if (editor == NULL || *editor == '\0')
 			    editor = VI;
 
-		    kill_line ();
-		    cmdbuf[0] = '+';
-		    scanstr (Currline - dlines <= 0 ? 1
-			     : Currline - (dlines + 1) / 2, &cmdbuf[1]);
-		    pr (editor); putchar (' ');
-		    pr (cmdbuf); putchar (' '); pr (fnames[fnum]);
-		    execute (filename, editor, editor, cmdbuf, fnames[fnum], 0);
+		    p = strrchr(editor, '/');
+		    if (p)
+			    p++;
+		    else
+			    p = editor;
+		    if (!strcmp(p, "vi") || !strcmp(p, "ex")) {
+			    sprintf(cmdbuf, "-c %d", n);
+			    split = 1;
+		    } else {
+			    sprintf(cmdbuf, "+%d", n);
+		    }
+
+		    kill_line();
+		    printf("%s %s %s", editor, cmdbuf, fnames[fnum]);
+		    if (split) {
+			    cmdbuf[2] = 0;
+			    execute(filename, editor, editor, cmdbuf,
+				    cmdbuf+3, fnames[fnum], (char *)0);
+		    } else
+			    execute(filename, editor, editor,
+				    cmdbuf, fnames[fnum], (char *)0);
 		    break;
 	    }
+	    /* fall through */
 	default:
 	    if (dum_opt) {
    		kill_line ();
 		if (Senter && Sexit) {
 		    my_putstring (Senter);
-		    promptlen = pr (_("[Press 'h' for instructions.]")) + (2 * soglitch);
+		    promptlen = printf(_("[Press 'h' for instructions.]"))
+			    + 2 * soglitch;
 		    my_putstring (Sexit);
 		}
 		else
-		    promptlen = pr (_("[Press 'h' for instructions.]"));
+		    promptlen = printf(_("[Press 'h' for instructions.]"));
 		fflush (stdout);
 	    }
 	    else
@@ -1394,9 +1394,9 @@ int colon (char *filename, int cmd, int nlines) {
 	case 'f':
 		kill_line ();
 		if (!no_intty)
-			promptlen = xprintf (_("\"%s\" line %d"), fnames[fnum], Currline);
+			promptlen = printf(_("\"%s\" line %d"), fnames[fnum], Currline);
 		else
-			promptlen = xprintf (_("[Not a file] line %d"), Currline);
+			promptlen = printf(_("[Not a file] line %d"), Currline);
 		fflush (stdout);
 		return (-1);
 	case 'n':
@@ -1463,11 +1463,11 @@ void do_shell (char *filename)
 	char *expanded;
 
 	kill_line ();
-	pr ("!");
+	putchar('!');
 	fflush (stdout);
 	promptlen = 1;
 	if (lastp)
-		pr (shell_line);
+		putsout(shell_line);
 	else {
 		ttyin (cmdbuf, sizeof(cmdbuf)-2, '!');
 		expanded = 0;
@@ -1480,16 +1480,16 @@ void do_shell (char *filename)
 			free(expanded);
 		}
 		if (rc < 0) {
-			errwrite(_("  Overflow\n"));
+			putserr(_("  Overflow\n"));
 			prompt (filename);
 			return;
 		} else if (rc > 0) {
 			kill_line ();
-			promptlen = xprintf ("!%s", shell_line);
+			promptlen = printf("!%s", shell_line);
 		}
 	}
 	fflush (stdout);
-	errwrite("\n");
+	putcerr('\n');
 	promptlen = 0;
 	shellp = 1;
 	execute (filename, shell, shell, "-c", shell_line, 0);
@@ -1524,10 +1524,10 @@ void search(char buf[], FILE *file, register int n)
 		if (--n == 0) {
 		    if (lncount > 3 || (lncount > 1 && no_intty))
 		    {
-			pr ("\n");
+			putchar('\n');
 			if (clreol)
 			    cleareol ();
-			pr(_("...skipping\n"));
+			putsout(_("...skipping\n"));
 		    }
 		    if (!no_intty) {
 			Currline -= (lncount >= 3 ? 3 : lncount);
@@ -1551,8 +1551,7 @@ void search(char buf[], FILE *file, register int n)
 			    else
 				doclear ();
 			}
-			pr (Line);
-			putchar ('\n');
+			puts(Line);
 		    }
 		    break;
 		}
@@ -1569,7 +1568,7 @@ void search(char buf[], FILE *file, register int n)
 	    Fseek (file, startline);
 	}
 	else {
-	    pr (_("\nPattern not found\n"));
+	    putsout(_("\nPattern not found\n"));
 	    end_it (0);
 	}
 	error (_("Pattern not found"));
@@ -1577,7 +1576,7 @@ void search(char buf[], FILE *file, register int n)
 }
 
 /*VARARGS2*/
-void execute (char * filename, char * cmd, ...)
+void execute (char *filename, char *cmd, ...)
 {
 	int id;
 	int n;
@@ -1619,7 +1618,7 @@ void execute (char * filename, char * cmd, ...)
 	    va_end(argp);
 	
 	    execvp (cmd, args);
-	    errwrite(_("exec failed\n"));
+	    putserr(_("exec failed\n"));
 	    exit (1);
 	}
 	if (id > 0) {
@@ -1633,9 +1632,9 @@ void execute (char * filename, char * cmd, ...)
 	    if (catch_susp)
 		signal(SIGTSTP, onsusp);
 	} else
-	    errwrite(_("can't fork\n"));
+	    putserr(_("can't fork\n"));
 	set_tty ();
-	pr ("------------------------\n");
+	puts("------------------------");
 	prompt (filename);
 }
 /*
@@ -1672,17 +1671,17 @@ void skipf (register int nskip)
     fnum += nskip;
     if (fnum < 0)
 	fnum = 0;
-    pr (_("\n...Skipping "));
-    pr ("\n");
+    puts(_("\n...Skipping "));
     if (clreol)
 	cleareol ();
-    pr (_("...Skipping "));
-    pr (nskip > 0 ? _("to file ") : _("back to file "));
-    pr (fnames[fnum]);
-    pr ("\n");
+    if (nskip > 0)
+	    putsout(_("...Skipping to file "));
+    else
+	    putsout(_("...Skipping back to file "));
+    puts(fnames[fnum]);
     if (clreol)
 	cleareol ();
-    pr ("\n");
+    putchar('\n');
     --fnum;
 }
 
@@ -1709,13 +1708,13 @@ retry:
 	     * Wait until we're in the foreground before we save the
 	     * the terminal modes.
 	     */
-	    if (ioctl(fileno(stdout), TIOCGPGRP, &tgrp) < 0) {
-		 perror("TIOCGPGRP");
-		 exit(1);
+	    if ((tgrp = tcgetpgrp(fileno(stdout))) < 0) {
+		perror("tcgetpgrp");
+		exit(1);
 	    }
 	    if (tgrp != getpgrp(0)) {
-		 kill(0, SIGTTOU);
-		 goto retry;
+		kill(0, SIGTTOU);
+		goto retry;
 	    }
 	}
 #endif
@@ -1815,10 +1814,10 @@ retry:
 }
 
 int readch () {
-	char c;
+	unsigned char c;
 
 	errno = 0;
-	if (read (2, &c, 1) <= 0) {
+	if (read (fileno(stderr), &c, 1) <= 0) {
 		if (errno != EINTR)
 			end_it(0);
 		else
@@ -1830,18 +1829,17 @@ int readch () {
 static char *BS = "\b";
 static char *BSB = "\b \b";
 static char *CARAT = "^";
-#define ERASEONECHAR \
+#define ERASEONECOLUMN \
     if (docrterase) \
-	errwrite(BSB); \
+	putserr(BSB); \
     else \
-	errwrite(BS);
+	putserr(BS);
 
-void ttyin (char buf[], register int nmax, char pchar) {
-    char *sp;
-    char c;
+void ttyin (unsigned char buf[], register int nmax, char pchar) {
+    unsigned char *sp;
+    int c;
     int slash = 0;
     int	maxlen;
-    char cbuf;
 
     sp = buf;
     maxlen = 0;
@@ -1853,12 +1851,59 @@ void ttyin (char buf[], register int nmax, char pchar) {
 	}
 	else if (((cc_t) c == otty.c_cc[VERASE]) && !slash) {
 	    if (sp > buf) {
-		--promptlen;
-		ERASEONECHAR
-		--sp;
+#ifdef ENABLE_WIDECHAR
+		if (MB_CUR_MAX > 1)
+		  {
+		    wchar_t wc;
+		    size_t pos = 0, mblength;
+		    mbstate_t state, state_bak;
+
+		    memset (&state, '\0', sizeof (mbstate_t));
+
+		    while (1) {
+			 state_bak = state;
+			 mblength = mbrtowc (&wc, buf + pos, sp - buf, &state);
+
+			 state = (mblength == (size_t)-2
+				 || mblength == (size_t)-1) ? state_bak : state;
+			 mblength = (mblength == (size_t)-2
+				     || mblength == (size_t)-1
+				     || mblength == 0) ? 1 : mblength;
+
+			 if (buf + pos + mblength >= sp)
+			 break;
+
+			 pos += mblength;
+		    }
+
+		    if (mblength == 1) {
+		      ERASEONECOLUMN
+		    }
+		    else {
+			int wc_width;
+			wc_width = wcwidth (wc);
+			wc_width = (wc_width < 1) ? 1 : wc_width;
+			while (wc_width--) {
+			    ERASEONECOLUMN
+			}
+		    }
+
+		    while (mblength--) {
+			--promptlen;
+			--sp;
+		    }
+		  }
+		else
+#endif
+		  {
+		    --promptlen;
+		    ERASEONECOLUMN
+		    --sp;
+		  }
+
 		if ((*sp < ' ' && *sp != '\n') || *sp == RUBOUT) {
 		    --promptlen;
-		    ERASEONECHAR
+		    ERASEONECOLUMN
 		}
 		continue;
 	    }
@@ -1880,7 +1925,7 @@ void ttyin (char buf[], register int nmax, char pchar) {
 		    erasep (1);
 		else if (docrtkill)
 		    while (promptlen-- > 1)
-			errwrite(BSB);
+			putserr(BSB);
 		promptlen = 1;
 	    }
 	    sp = buf;
@@ -1889,7 +1934,7 @@ void ttyin (char buf[], register int nmax, char pchar) {
 	}
 	if (slash && ((cc_t) c == otty.c_cc[VKILL]
 		   || (cc_t) c == otty.c_cc[VERASE])) {
-	    ERASEONECHAR
+	    ERASEONECOLUMN
 	    --sp;
 	}
 	if (c != '\\')
@@ -1897,12 +1942,11 @@ void ttyin (char buf[], register int nmax, char pchar) {
 	*sp++ = c;
 	if ((c < ' ' && c != '\n' && c != ESC) || c == RUBOUT) {
 	    c += (c == RUBOUT) ? -0100 : 0100;
-	    errwrite(CARAT);
+	    putserr(CARAT);
 	    promptlen++;
 	}
-	cbuf = c;
 	if (c != '\n' && c != ESC) {
-	    errwrite1(&cbuf);
+	    putcerr(c);
 	    promptlen++;
 	}
 	else
@@ -1974,28 +2018,15 @@ int expand (char **outbuf, char *inbuf) {
 }
 
 void show (char c) {
-    char cbuf;
-
     if ((c < ' ' && c != '\n' && c != ESC) || c == RUBOUT) {
 	c += (c == RUBOUT) ? -0100 : 0100;
-	errwrite(CARAT);
+	putserr(CARAT);
 	promptlen++;
     }
-    cbuf = c;
-    errwrite1(&cbuf);
+    putcerr(c);
     promptlen++;
 }
 
-void errwrite (char *txt)
-{
-    write (fileno(stderr), txt, strlen(txt));
-}
-
-void errwrite1 (char *sym)
-{
-    write (fileno(stderr), sym, 1);
-}
-
 void error (char *mess)
 {
     if (clreol)
@@ -2005,11 +2036,11 @@ void error (char *mess)
     promptlen += strlen (mess);
     if (Senter && Sexit) {
 	my_putstring (Senter);
-	pr(mess);
+	putsout(mess);
 	my_putstring (Sexit);
     }
     else
-	pr (mess);
+	putsout(mess);
     fflush(stdout);
     errors++;
     siglongjmp (restore, 1);
@@ -2025,7 +2056,7 @@ void set_tty () {
 
 static int
 ourputch(int c) {
-    return putc(c, stdout);
+	return putc(c, stdout);
 }
 
 void
@@ -2033,7 +2064,7 @@ reset_tty () {
     if (no_tty)
 	return;
     if (pstate) {
-	tputs(ULexit, 1, ourputch);
+	tputs(ULexit, 1, ourputch);	/* putchar - if that isnt a macro */
 	fflush(stdout);
 	pstate = 0;
     }
